-module(akh_lexer).
-include("akhamoth.hrl").
-export([
    new/1,
    next/1
]).

-type token_kind() ::
    binary()
    | '{'
    | '}'
    | '['
    | ']'
    | '('
    | ')'
    | ','
    | ';'
    | ':'
    | '.'
    | '->'
    | '+'.

-type token() :: {token_kind(), non_neg_integer() | inserted}.

-define(is_id(C), (C >= $A andalso C =< $Z); (C >= $a andalso C =< $z); C =:= $_).

-record(lexer, {
    source :: binary(),
    offset = 0 :: non_neg_integer()
}).

-spec new(binary()) -> #lexer{}.
new(Source) -> #lexer{source = Source}.

-spec next(#lexer{}) -> none | {ok, token(), #lexer{}}.
next(#lexer{source = <<$\n, Rest/bytes>>, offset = Offset} = Lexer) ->
    next(Lexer#lexer{source = Rest, offset = Offset + 1});
next(#lexer{source = <<C, _/bytes>>} = Lexer) when ?is_id(C) ->
    lex_id(Lexer, 1);
next(#lexer{source = <<>>}) ->
    none.

-spec lex_id(#lexer{}, pos_integer()) -> {ok, token(), #lexer{}}.
lex_id(#lexer{source = Source, offset = Offset} = Lexer, Len) when Len < byte_size(Source) ->
    C = binary:at(Source, Len),
    if
        ?is_id(C) ->
            lex_id(Lexer, Len + 1);
        true ->
            {Id, Rest} = split_binary(Source, Len),
            {ok, {Id, Offset}, Lexer#lexer{source = Rest, offset = Offset + Len}}
    end;
lex_id(#lexer{source = Source, offset = Offset} = Lexer, Len) ->
    {ok, {Source, Offset}, Lexer#lexer{source = <<>>, offset = Offset + Len}}.
